from .urls import REQUEST_ADDRESS
from .urls import (TCGA_ESTIMATED_SIZE_FIELD,
                   TCGA_SUBMISSION_TIME_FIELD,
                   TCGA_TICKET_ID_FIELD,
                   TCGA_STATUS_CHECK_URL_FIELD)

import requests
import logging
import time
import os
import hashlib
import json

from appdirs import user_data_dir

PYTCGA_BASE_DIRECTORY = user_data_dir("pytcga", version="0.1")

def check_if_exists_cached_file(output_file_name):
    return os.path.exists(
        os.path.join(PYTCGA_BASE_DIRECTORY, output_file_name)
    )


def tcga_request(disease,
                  center=None,
                  level=None,
                  platform=None,
                  platformType=None,
                  sample_list=None,
                  consolidateFiles='true',
                  flattenDir='true',
                  cache=True,
                  wait_time=30):

    # All disease codes must be upper-case.
    disease = disease.upper()

    filter_parameters = {
        'disease': disease,
        'center': center,
        'level': level,
        'platform': platform,
        'platformType': platformType,
        'sampleList': sample_list,
        'flattenDir': flattenDir,
        'consolidateFiles': consolidateFiles
    }

    # Save hash of request parameters
    request_id = hashlib.md5(
                        json.dumps(filter_parameters,
                                   sort_keys=True).encode('utf-8')
                    ).hexdigest()

    # Create an output tar file with that ID
    output_file_name = request_id + '.tar'

    # If using the cache, check if the file already exists
    if cache:
        archive_path = os.path.join(PYTCGA_BASE_DIRECTORY, output_file_name)

        if os.path.exists(archive_path):
            return archive_path

    (ticket_id, status_url) = create_tcga_request(filter_parameters)
    return check_and_retrieve_archive(
                                status_url, 
                                archive_file_name=output_file_name, 
                                wait_time=wait_time)


def create_tcga_filter_request(disease,
                              center=None,
                              level=None,
                              platform=None,
                              platformType=None,
                              sample_list=None,
                              consolidateFiles='true',
                              flattenDir='true'):
    """Create a web service request for TCGA

    Parameters
    ----------
    disease : str
        TCGA disease code {'LUAD', 'BRCA', 'BLAC',...}
    center : str
        TCGA center code {'BI', ...}
    level : str, optional
        Level 1, 2 or 3
    platform : str
        Platform {'Automated Mutation Calling', 'Curated Mutation Calling', ...}
    platformType : str
        Platform type {'Somatic Mutations', ...}
    sample_list : str, optional
        List of specific TCGA IDs to request
    consolidateFiles : str, optional
        Combine files of the same type, 'true' or 'false'
    flattenDir : str, optional
        Flatten the directory structure, 'true' or 'false'

    Raises
    ------
    ValueError
        A platform is required for the web service

    Returns
    -------
    (ticket_id, status_url) : (str, str)
        A pair of the ticket_id created and status url
    """
    if platform is None:
        raise ValueError(
            """Platform must be specfied."""
        )

    # All disease codes must be upper-case.
    disease = disease.upper()

    filter_parameters = {
        'disease': disease,
        'center': center,
        'level': level,
        'platform': platform,
        'platformType': platformType,
        'sampleList': sample_list,
        'flattenDir': flattenDir,
        'consolidateFiles': consolidateFiles
    }

    return create_tcga_request(filter_parameters)

def create_tcga_request(filter_parameters):
    """Builds a webservice request from a dictionary of parameters

    Parameters
    ----------
    filter_parameters : dict
        Dictionary of filter parameters for TCGA web service request

    Returns
    -------
    (ticket_id, status_url) : (str, str)
        A pair of the ticket_id created and status url
    """
    response = requests.get(REQUEST_ADDRESS, params=filter_parameters)

    logging.debug("Request has status code {}".format(response.status_code))

    if response.status_code == 200:
        parsed_response = response.json()

        ticket_id = parsed_response[TCGA_TICKET_ID_FIELD]
        submission_time = parsed_response[TCGA_SUBMISSION_TIME_FIELD]
        estimated_size = parsed_response[TCGA_ESTIMATED_SIZE_FIELD]
        status_url = parsed_response[TCGA_STATUS_CHECK_URL_FIELD]

        logging.info(
            'Request received at {} for ticket {}, estimated size {}'.format(
                submission_time, 
                ticket_id,
                estimated_size
            )
        )

        logging.info(
            'Track ticket {} at {}'.format( 
                ticket_id,
                status_url
            )
        )
    else:
        raise ValueError('Request {} failed, \n{}'.format(response.url, response.text))

    return (ticket_id, status_url)

def retrieve_ticket_status(status_url):
    """Extract the current status at the tracking/status url

    Parameters
    ----------
    status_url : str
        Tracking URL for a submitted data request

    Returns
    -------
    job_status : str
        Current status {'OK', 'Accepted', ...}
    """
    tracking_response = requests.get(status_url)
    tracking_response_parsed = tracking_response.json()
    job_status = tracking_response_parsed['job-status']

    return job_status

def retrieve_archive(archive_url,
                    output_file_name,
                    block_size=1024):
    """Download the archive from given URL into the output file

    Parameters
    ----------
    archive_url : str
        TCGA URL to the created archive
    output_file_name : str
        Filename to save the archive

    Returns
    -------
    archive_path : str
        Full path to the downloaded archive
    """
    archive_path = os.path.join(PYTCGA_BASE_DIRECTORY, output_file_name)
    logging.info('Saving request to {}'.format(archive_path))

    with open(archive_path, 'wb') as archive:
        archive_response = requests.get(archive_url, stream=True)

        for block in archive_response.iter_content(block_size):
            archive.write(block)

    return archive_path

def check_and_retrieve_archive(status_url,
                               archive_file_name,
                               wait_time=None):
    """Checks the status URL from TCGA and attempts to download the archive

    Returns None if no archive exists and did not re-poll

    Parameters
    ----------
    status_url : str
        URL to check the status of the data request
    archive_file_name : str
        File name for archive from TCGA data request
    wait_time : int, optional
        Wait-time (in seconds) before polling service for data

    Returns
    -------
    archive_path : str
        Return a path to the downloaded archive
        Returns None if no archive exists and did not re-poll
    """
    job_status = retrieve_ticket_status(status_url)
    if job_status['status-message'] == 'OK':
        archive_url = job_status['archive-url']
        return retrieve_archive(archive_url, archive_file_name)

    if wait_time:
        while(True):
            time.sleep(wait_time)
            job_status = retrieve_ticket_status(status_url)

            if job_status['status-message'] == 'OK':
                archive_url = job_status['archive-url']
                archive_path = retrieve_archive(archive_url, archive_file_name)
                return archive_path

    return None


import os
import logging
import tarfile
import pandas as pd 

from pytcga.tcga_requests import tcga_request
from pytcga.tcga_clinical import load_clinical_data

def prefetch_mutation_data(disease_code,
                          wait_time=30,
                          cache=True,):

    archive_path = tcga_request(disease=disease_code,
                               level='2',
                               center='BI',
                               platformType='Somatic Mutations',
                               platform='Automated Mutation Calling',
                               wait_time=wait_time,
                               cache=cache)

    return archive_path

def load_mutation_data(disease_code,
                       with_clinical=False,
                       variant_type='all',
                       wait_time=30):

    archive_path = prefetch_mutation_data(disease_code,
                                      wait_time=wait_time,
                                      cache=True)

    # Unpack tar file
    archive = tarfile.open(archive_path)

    result_dir = os.path.join(os.path.dirname(archive_path), disease_code, 'mutations')
    if not os.path.exists(result_dir):
        os.makedirs(result_dir)

    archive.extractall(path=result_dir)

    # Filter to MAF files
    maf_files = [f 
                  for f in os.listdir(result_dir) 
                  if f.endswith('.maf')]

    mutation_df = pd.concat([pd.read_csv(os.path.join(result_dir, maf_file),
                                    sep='\t', 
                                    na_values='[Not Available]') 
                    for maf_file in maf_files])

    # Expand out the TCGA barcode to retrieve the TCGA ID
    tcga_info = mutation_df['Tumor_Sample_Barcode'].str.rsplit('-', n=4, expand=True)
    tcga_info.columns = ['TCGA_ID', 'SampleID', 'PortionID', 'PlateID', 'CenterID']

    mutations = mutation_df.join(tcga_info, how='left')

    if variant_type != 'all':
        if variant_type == 'indel':
            mutations = mutations[
                            (mutations['Variant_Type'] == 'INS') |
                            (mutations['Variant_Type'] == 'DEL')
                        ]          
        else:     
            mutations = mutations[mutations['Variant_Type'] == variant_type]

    logging.info("Loaded {} mutations for {} tumors from {} patients".format(
                    len(mutations),
                    mutations['Tumor_Sample_Barcode'].nunique(),
                    mutations['TCGA_ID'].nunique()
                )
    )

    if with_clinical:
        patient_data_df = load_clinical_data(disease_code)
        merged = mutations.merge(patient_data_df,
                            how='outer',
                            left_on='TCGA_ID', 
                            right_on='bcr_patient_barcode')

        logging.info("Patients: {}, Tumor Samples: {}, Mutations {}".format(
                    merged['bcr_patient_barcode'].nunique(),
                    merged['Tumor_Sample_Barcode'].nunique(),
                    len(merged[~merged['TCGA_ID'].isnull()])
                )
        )
        return merged
    else: 
        return mutations



import os
import tarfile
import pandas as pd

from pytcga.tcga_requests import tcga_request
from pytcga.tcga_clinical import load_clinical_data

def prefetch_rnaseq_data(disease_code,
                        wait_time=30,
                        cache=True,):

    archive_path = tcga_request(disease=disease_code,
                               level='3',
                               center='7',
                               platformType='RNASeqV2',
                               platform='IlluminaHiSeq_RNASeqV2',
                               wait_time=30)

    return archive_path


def load_rnaseq_data(disease_code,
                     with_clinical=False,
                     wait_time=30):
    # Fetch RNA data
    archive_path = prefetch_rnaseq_data(disease_code)

    # Unpack tar file
    archive = tarfile.open(archive_path)
    result_dir = os.path.join(os.path.dirname(archive_path), disease_code, 'gene_expression')

    if not os.path.exists(result_dir):
        os.makedirs(result_dir)

    gene_quantification_files = set(el for el in archive if 'genes.normalized_results' in el.name)
    if len(gene_quantification_files.difference(os.listdir(result_dir))) > 0:
        archive.extractall(members=gene_quantification_files, path=result_dir)
        archive.extract('FILE_SAMPLE_MAP.txt', path=result_dir)

    gene_quantification_files = [el.name for el in gene_quantification_files]

    # Load map from samples to RNA files
    rna_file_sample_map = pd.read_csv(os.path.join(result_dir, 'FILE_SAMPLE_MAP.txt'), sep='\t')
    rna_file_sample_map_id_split = rna_file_sample_map['barcode(s)'].str.rsplit('-', n=4, expand=True)
    rna_file_sample_map_id_split.columns = ['TCGA_ID', 'SampleID', 'PortionID', 'PlateID', 'CenterID']

    rna_file_sample_map = rna_file_sample_map.join(rna_file_sample_map_id_split)

    gene_filter = rna_file_sample_map['filename'].str.contains('genes.normalized_results')
    gene_rna_file_sample_map = rna_file_sample_map[gene_filter]

    rna_dfs = []
    for (f, sample) in list(zip(gene_rna_file_sample_map['filename'], gene_rna_file_sample_map['TCGA_ID'])):
        sample_rna_df = pd.read_csv(os.path.join(result_dir, f), sep='\t')
        sample_rna_df['TCGA_ID'] = sample

        sample_rna_df['gene_name'] = sample_rna_df.gene_id.str.split('|').str.get(0)
        rna_dfs.append(sample_rna_df)

    rna_df = pd.concat(rna_dfs).merge(rna_file_sample_map)

    if with_clinical:
        patient_data_df = load_clinical_data(disease_code)
        merged = rna_df.merge(patient_data_df,
                              how='outer',
                              left_on='TCGA_ID', 
                              right_on='bcr_patient_barcode')

        return merged
    else: 
        return rna_df



clinical_data_dictionary = {
   
    'tobacco_smoking_history':
        {
            '1': 'Lifelong Non-smoker',
            '2': 'Current smoker',
            '3': 'Current reformed smoker for > 15 years',
            '4': 'Current reformed smoker for < or = 15 years',
            '5': 'Current Reformed Smoker, Duration Not Specified',

        }
}


import os
import logging
import requests
from bs4 import BeautifulSoup
import pandas as pd

from .tcga_requests import PYTCGA_BASE_DIRECTORY
from .clinical_data_dictionary import clinical_data_dictionary

TCGA_CLINICAL_URL = "https://tcga-data.nci.nih.gov/tcgafiles/ftp_auth/distro_ftpusers/anonymous/tumor/{}/bcr/biotab/clin/"

PATIENT_DATA_FILE_CODE = 'clinical_patient'
def request_clinical_data(disease_code,
                  cache=True,
                  block_size=1024):
    """Downloads TCGA public clinical data from the TCGA FTP site
    
    Parameters
    ----------
    disease_code : str
        TCGA disease type, i.e. 'LUAD', 'BLCA', 'BRCA' etc.
    cache : bool, optional
        Whether to cache the results of the request
    block_size : int, optional
        Block size for file downloads
    
    Returns
    -------
    patient_data_path : str
        Path to TCGA patient data file after downloading
    """
    # Create directory to save clinical data
    disease_code_dir = os.path.join(PYTCGA_BASE_DIRECTORY, disease_code)

    if cache and os.path.exists(disease_code_dir):
        patient_data_file = [f for f in os.listdir(disease_code_dir) if PATIENT_DATA_FILE_CODE in f]

        if len(patient_data_file) == 1:
            return os.path.join(disease_code_dir, patient_data_file[0])

    if not os.path.exists(disease_code_dir):
        os.mkdir(disease_code_dir)

    clinical_data_directory = TCGA_CLINICAL_URL.format(disease_code.lower())
    r = requests.get(clinical_data_directory)
    soup = BeautifulSoup(r.content)

    # Retrieve list of files and filter to txt files
    file_links = [link.get('href') 
                    for link in soup.find_all('a')]
    clinical_files = [link for link in file_links if link.endswith('.txt')]

    # Download all clinical data files
    patient_data_path = None
    for clinical_file in clinical_files:
        output_file = os.path.join(disease_code_dir, clinical_file)
        logging.debug('Saving {} clinical data request to {}'.format(clinical_file, output_file))

        if PATIENT_DATA_FILE_CODE in output_file:
            patient_data_path = output_file

        with open(output_file, 'wb') as archive:
            archive_response = requests.get(clinical_data_directory + '/' + clinical_file, stream=True)

            for block in archive_response.iter_content(block_size):
                archive.write(block)

    return patient_data_path


def load_clinical_data(disease_code, recode_columns=True):
    """Downloads and loads the TCGA clinical data into a Pandas dataframe

    Parameters
    ----------
    disease_code : str
        TCGA disease type, i.e. 'LUAD', 'BLCA', 'BRCA' etc.

    Returns
    -------
    patient_data_df : Dataframe
        Returns a Pandas dataframe with the patient data
    """
    patient_data_path = request_clinical_data(disease_code, cache=True)

    columns = pd.read_csv(patient_data_path,
                            sep='\t',
                            skiprows=1,
                            nrows=10).columns

    patient_data_df = pd.read_csv(patient_data_path,
                    sep='\t', 
                    skiprows=2,
                    header=0,
                    names=columns,
                    na_values='[Not Available]')

    if recode_columns:
        for column in clinical_data_dictionary:
            if column in patient_data_df.columns:
                patient_data_df[column] = patient_data_df[column].map(
                        lambda val: clinical_data_dictionary[column].get(val, val)
                    )

    logging.info("Loaded {} rows of clinical data from {} patients".format(
            len(patient_data_df),
            patient_data_df['bcr_patient_barcode'].nunique()
        )
    )

    return patient_data_df


from __future__ import print_function
import os

from setuptools import setup, find_packages

readme_dir = os.path.dirname(__file__)
readme_filename = os.path.join(readme_dir, 'README.md')

try:
    with open(readme_filename, 'r') as f:
        readme = f.read()
except:
    readme = ""

try:
    import pypandoc
    readme = pypandoc.convert(readme, to='rst', format='md')
except:
    print(
        "Conversion of long_description from MD to reStructuredText failed...")


if __name__ == '__main__':
    setup(
        name='pytcga',
        version="0.0.3",
        description="Store and query public TCGA data",
        author="Arun Ahuja",
        author_email="aahuja11 {at} gmail {dot}",
        url="https://github.com/arahuja/pytcga",
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Environment :: Console',
            'Operating System :: OS Independent',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
        ],
        install_requires=[
            'pandas >=0.13.1',
            'nose >=1.3.6',
            'beautifulsoup4',
            'requests',
        ],
        long_description=readme,
        packages=find_packages(exclude=["test", "tests"]),
    )



TCGA_NIH_REQUEST_ADDRESS = "http://tcga-data.nci.nih.gov/tcga/damws/jobprocess/json"

TCGA_TICKET_ID_FIELD = 'ticket'
TCGA_SUBMISSION_TIME_FIELD = 'submission-time'
TCGA_ESTIMATED_SIZE_FIELD = 'estimated-size'
TCGA_STATUS_CHECK_URL_FIELD = 'status-check-url'

REQUEST_ADDRESS = TCGA_NIH_REQUEST_ADDRESS
JOB_STATUS_ADDRESS = "http://tcga-data.nci.nih.gov/tcga/damws/jobprocess/json/ticket/"
